import os
import string
from collections import OrderedDict
from dataclasses import dataclass
from distutils.version import LooseVersion
from functools import reduce as _reduce
from os.path import exists, join, basename
from typing import Dict, Iterable, List, Set, Sequence, Tuple

from robotpy_installer.errors import OpkgError
from robotpy_installer.utils import _urlretrieve, md5sum

Package = OrderedDict


@dataclass
class Feed:
    url: str
    db_fname: str
    pkgs: Dict[str, List[Package]]
    loaded: bool


class OpkgRepo(object):
    """Simplistic OPkg Manager"""

    sys_packages = ["libc6"]

    def __init__(self, opkg_cache, arch: str, ssl_context):
        self.feeds: List[Feed] = []
        self.opkg_cache = opkg_cache
        self.arch = arch
        self.ssl_context = ssl_context
        if not exists(self.opkg_cache):
            os.makedirs(self.opkg_cache)
        self.pkg_dbs = join(self.opkg_cache, "Packages")
        if not exists(self.pkg_dbs):
            os.makedirs(self.pkg_dbs)

    def add_feed(self, url: str) -> None:
        # Snippet from https://gist.github.com/seanh/93666
        valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        safe_url = "".join(c for c in url if c in valid_chars)
        safe_url = safe_url.replace(" ", "_")
        feed = Feed(
            url,
            db_fname=join(self.pkg_dbs, safe_url),
            pkgs=OrderedDict(),
            loaded=False,
        )
        if exists(feed.db_fname):
            self.load_package_db(feed)
            feed.loaded = True

        self.feeds.append(feed)

    def update_packages(self) -> None:
        for feed in self.feeds:
            pkgurl = feed.url + "/Packages"
            _urlretrieve(pkgurl, feed.db_fname, True, self.ssl_context)
            self.load_package_db(feed)

    def load_package_db(self, feed: Feed) -> None:
        # dictionary of lists of packages sorted by version
        pkg: Package = OrderedDict()
        with open(feed.db_fname, "r", encoding="utf-8") as fp:
            for line in fp.readlines():
                line = line.strip()
                if len(line) == 0:
                    self._add_pkg(pkg, feed)
                    pkg = OrderedDict()
                else:
                    if ":" in line:
                        k, v = [i.strip() for i in line.split(":", 1)]
                        if k == "Version":
                            pkg[k] = LooseVersion(v)
                        else:
                            pkg[k] = v

        self._add_pkg(pkg, feed)

        # Finally, make sure all the packages are sorted by version
        for pkglist in feed.pkgs.values():
            pkglist.sort(key=lambda p: p["Version"])

    def _add_pkg(self, pkg: Package, feed: Feed) -> None:
        if len(pkg) == 0 or pkg.get("Architecture", None) != self.arch:
            return
        # Add download url and fname
        if "Filename" in pkg:
            pkg["url"] = "/".join((feed.url, pkg["Filename"]))

        # Only retain one version of a package
        pkgs = feed.pkgs.setdefault(pkg["Package"], [])
        for old_pkg in pkgs:
            if old_pkg["Version"] == pkg["Version"]:
                old_pkg.clear()
                old_pkg.update(pkg)
                break
        else:
            pkgs.append(pkg)

    def get_pkginfo(self, name: str):
        loaded = False
        for feed in self.feeds:
            loaded = loaded or feed.loaded
            if name in feed.pkgs:
                return feed.pkgs[name][-1]

        if loaded:
            msg = "Package %s is not in the package list (did you misspell it?)" % name
        else:
            msg = "There are no package lists, did you download %s yet?" % name

        raise OpkgError(msg)

    def _get_pkg_fname(self, pkg: Package) -> str:
        return join(self.opkg_cache, basename(pkg["Filename"]))

    def _get_pkg_deps(self, name: str) -> Set[str]:
        info = self.get_pkginfo(name)
        if "Depends" in info:
            return {
                dep
                for dep in [
                    dep.strip().split(" ", 1)[0] for dep in info["Depends"].split(",")
                ]
                if dep not in self.sys_packages
            }
        return set()

    def get_cached_pkg(self, name: str) -> Tuple:
        """Returns the pkg, filename of a cached package"""
        pkg = self.get_pkginfo(name)
        fname = self._get_pkg_fname(pkg)

        if not exists(fname):
            raise OpkgError("Package '%s' has not been downloaded" % name)

        if not md5sum(fname) == pkg["MD5Sum"]:
            raise OpkgError("md5sum of package '%s' md5sum does not match" % name)

        return pkg, fname

    def resolve_pkg_deps(self, packages: Sequence[str]) -> List[str]:
        """Given a list of package(s) desired to be installed, topologically
        sorts them by dependencies and returns an ordered list of packages"""

        pkgs = {}
        packages = list(packages)

        for pkg in packages:
            if pkg in pkgs:
                continue
            deps = self._get_pkg_deps(pkg)
            pkgs[pkg] = deps
            packages.extend(deps)

        retval: List[str] = []
        for results in self._toposort(pkgs):
            retval.extend(results)

        return retval

    @classmethod
    def _toposort(cls, data: Dict[str, Set[str]]) -> Iterable[Set[str]]:
        # Copied from https://bitbucket.org/ericvsmith/toposort/src/25b5894c4229cb888f77cf0c077c05e2464446ac/toposort.py?at=default
        # -> Apache 2.0 license, Copyright 2014 True Blade Systems, Inc.

        # Special case empty input.
        if len(data) == 0:
            return

        # Copy the input so as to leave it unmodified.
        data = data.copy()

        # Ignore self dependencies.
        for k, v in data.items():
            v.discard(k)
        # Find all items that don't depend on anything.
        extra_items_in_deps = _reduce(set.union, data.values()) - set(data.keys())
        # Add empty dependences where needed.
        data.update({item: set() for item in extra_items_in_deps})
        while True:
            ordered = {item for item, dep in data.items() if len(dep) == 0}
            if not ordered:
                break
            yield ordered
            data = {
                item: (dep - ordered)
                for item, dep in data.items()
                if item not in ordered
            }
        if len(data) != 0:
            yield cls._modified_dfs(data)

    @staticmethod
    def _modified_dfs(nodes: Dict[str, Set[str]]):
        # this is a modified depth first search that does a best effort at
        # a topological sort, but ignores cycles and keeps going on despite
        # that. Only used if the topological sort fails.
        retval = []
        visited = set()

        def _visit(n):
            if n in visited:
                return

            visited.add(n)
            for m in nodes[n]:
                _visit(m)

            retval.append(n)

        for item in nodes:
            _visit(item)

        return retval

    def download(self, name: str) -> str:
        pkg = self.get_pkginfo(name)
        fname = self._get_pkg_fname(pkg)

        # Only download it if necessary
        if not exists(fname) or not md5sum(fname) == pkg["MD5Sum"]:
            _urlretrieve(pkg["url"], fname, True, self.ssl_context)
        # Validate it
        if md5sum(fname) != pkg["MD5Sum"]:
            raise OpkgError("Downloaded package for %s md5sum does not match" % name)

        return fname

    def load_opkg_from_req(self, *files: str) -> List[str]:
        """
        Pull the list of opkgs from a requirements.txt-like file
        """
        opkgs = []
        # Loop through the passed in files to support multiple requirements files
        for file in files:
            with open(file, "r") as f:
                for row in f:
                    # Ignore commented lines and empty lines
                    stripped = row.strip()
                    if stripped and not stripped.startswith("#"):
                        # Add the package to the list of packages (and remove leading and trailing whitespace)
                        opkgs.append(stripped)
        return opkgs



import contextlib
import hashlib
import json
import logging
import socket
import sys
import urllib.request
from os.path import exists

from robotpy_installer import __version__
from robotpy_installer.errors import Error

logger = logging.getLogger("robotpy.installer")

_useragent = "robotpy-installer/%s" % __version__


def md5sum(fname):
    md5 = hashlib.md5()
    with open(fname, "rb") as fp:
        buf = fp.read(65536)
        while len(buf) > 0:
            md5.update(buf)
            buf = fp.read(65536)
    return md5.hexdigest()


def _urlretrieve(url, fname, cache, ssl_context):
    # Get it
    print("Downloading", url)

    # Save bandwidth! Use stored metadata to prevent re-downloading
    # stuff we already have
    last_modified = None
    etag = None
    cache_fname = None

    if cache:
        cache_fname = fname + ".jmd"
        if exists(fname) and exists(cache_fname):
            try:
                with open(cache_fname) as fp:
                    md = json.load(fp)
                if md5sum(fname) == md["md5"]:
                    etag = md.get("etag")
                    last_modified = md.get("last-modified")
            except Exception:
                pass

    blocksize = 1024 * 8

    def _reporthook(read, totalsize):
        if totalsize > 0:
            percent = min(int(read * 100 / totalsize), 100)
            sys.stdout.write("\r%02d%%" % percent)
        else:
            sys.stdout.write("\r%dbytes" % read)
        sys.stdout.flush()

    try:
        # adapted from urlretrieve source
        headers = {"User-Agent": _useragent}
        if last_modified:
            headers["If-Modified-Since"] = last_modified
        if etag:
            headers["If-None-Match"] = etag

        req = urllib.request.Request(url, headers=headers)

        with contextlib.closing(
            urllib.request.urlopen(req, context=ssl_context)
        ) as rfp:
            headers = rfp.info()

            with open(fname, "wb") as fp:
                # Deal with header stuff
                size = -1
                read = 0
                if "content-length" in headers:
                    size = int(headers["Content-Length"])

                while True:
                    block = rfp.read(blocksize)
                    if not block:
                        break
                    read += len(block)
                    fp.write(block)
                    _reporthook(read, size)

        if size >= 0 and read < size:
            raise ValueError("Only retrieved %s of %s bytes" % (read, size))

        # If we received info from the server, cache it
        if cache_fname:
            md = {}
            if "etag" in headers:
                md["etag"] = headers["ETag"]
            if "last-modified" in headers:
                md["last-modified"] = headers["Last-Modified"]
            if md:
                md["md5"] = md5sum(fname)
                with open(cache_fname, "w") as fp:
                    json.dump(md, fp)
    except urllib.error.HTTPError as e:
        if e.code == 304:
            sys.stdout.write("Not modified")
        else:
            raise
    except Exception as e:
        if "certificate verify failed" in str(e) and sys.platform == "darwin":
            pyver = ".".join(map(str, sys.version_info[:2]))
            msg = (
                "SSL certificates are not installed! Run /Applications/Python %s/Install Certificates.command to fix this"
                % pyver
            )
            raise Exception(msg) from e
        else:
            raise e
    sys.stdout.write("\n")


def _resolve_addr(hostname):
    try:
        logger.debug("Looking up hostname '%s'...", hostname)
        # Note: Windows will never return a SOCK_STREAM address if you don't explicitly
        #       ask for it here. macOS and Linux always return all types, but filter it
        #       for us if we specify it here, so it doesn't hurt to specify it.
        addrs = socket.getaddrinfo(hostname, None, 0, socket.SOCK_STREAM)
    except socket.gaierror as e:
        raise Error("Could not find robot at %s" % hostname) from e

    # Sort the address by family.
    # Lucky for us, the family type is the first element of the tuple, and it's an enumerated type with
    # AF_INET=2 (IPv4) and AF_INET6=23 (IPv6), so sorting them will provide us with the AF_INET address first.
    addrs.sort()

    # pick the first address that is sock_stream
    # AF_INET sockaddr tuple:  (address, port)
    # AF_INET6 sockaddr tuple: (address, port, flow info, scope id)
    for _, socktype, _, _, sockaddr in addrs:
        if socktype == socket.SOCK_STREAM:
            ip = sockaddr[
                0
            ]  # The address if the first tuple element for both AF_INET and AF_INET6
            logger.debug("-> Found %s at %s" % (hostname, ip))
            return ip

    raise Error("Could not find robot at %s" % hostname)



import configparser
import io
import logging
import re
import os
from os.path import exists, join, expanduser, split as splitpath
from pathlib import PurePath, PurePosixPath
import socket
import typing


import paramiko

from robotpy_installer.errors import SshExecError, Error
from robotpy_installer.robotfinder import RobotFinder
from robotpy_installer.utils import _resolve_addr

logger = logging.getLogger("robotpy.installer")


class SuppressKeyPolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        return


class SshExecResult(typing.NamedTuple):
    returncode: int
    stdout: typing.Optional[str]


class SshController:
    """
    Use this to execute commands on a roboRIO in a cross platform manner

    ::

        with SshController(hostname, username, password) as controller:
            controller.exec_cmd("cat /etc/lsb-release", print_output=True)

    """

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        conn: typing.Optional[socket.socket],
    ):
        self.username = username
        self.password = password
        self.hostname = hostname
        self.conn = conn

        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(SuppressKeyPolicy)

    def __enter__(self):
        self.client.connect(
            self.hostname,
            username=self.username,
            password=self.password,
            allow_agent=False,
            look_for_keys=False,
            sock=self.conn,
        )
        return self

    def __exit__(self, *args):
        self.client.close()

    def exec_cmd(
        self,
        cmd: str,
        *,
        check: bool = False,
        get_output: bool = False,
        print_output: bool = False,
    ) -> SshExecResult:
        output = None
        buffer = io.StringIO()

        transport = self.client.get_transport()
        assert transport is not None

        with transport.open_session() as channel:
            channel.set_combine_stderr(True)
            channel.exec_command(cmd)

            with channel.makefile("r") as stdout:
                for line in stdout:
                    if get_output:
                        buffer.write(line)
                    if print_output:
                        print(line, end="")

            retval = channel.recv_exit_status()

        if check and retval != 0:
            raise SshExecError(
                "Command '%s' returned non-zero error status %s" % (cmd, retval),
                retval,
            )
        elif get_output:
            output = buffer.getvalue()

        return SshExecResult(retval, output)

    def check_output(self, cmd: str, *, print_output: bool = False) -> str:
        result = self.exec_cmd(
            cmd,
            check=True,
            get_output=True,
            print_output=print_output,
        )
        assert result.stdout is not None
        return result.stdout

    def sftp(self, local_path, remote_path, mkdir=True):
        # from https://gist.github.com/johnfink8/2190472
        oldcwd = os.getcwd()
        sftp = self.client.open_sftp()
        try:
            remote_path = PurePosixPath(remote_path)
            parent, child = splitpath(local_path)
            os.chdir(parent)
            for d, _, files in os.walk(child):
                d = PurePath(d)
                try:
                    remote_dir = remote_path / d
                    print("make", remote_dir)
                    if not mkdir:
                        # skip first mkdir
                        mkdir = True
                    else:
                        sftp.mkdir(str(remote_dir))
                except:
                    raise
                for fname in files:
                    local_fname = d / fname
                    remote_fname = remote_dir / fname
                    print(local_fname.relative_to(child), "->", remote_fname)
                    sftp.put(str(local_fname), str(remote_fname))
        finally:
            os.chdir(oldcwd)
            sftp.close()


def ssh_from_cfg(
    cfg_filename: typing.Union[str, os.PathLike],
    username: str,
    password: str,
    hostname: typing.Optional[str] = None,
    no_resolve=False,
):
    # hostname can be a team number or an ip / hostname

    dirty = True
    cfg = configparser.ConfigParser()
    cfg.setdefault("auth", {})

    if exists(cfg_filename):
        logger.info("-> using existing config at '%s'", str(cfg_filename))
        cfg.read(cfg_filename)
        dirty = False

    if hostname is not None:
        dirty = True
        cfg["auth"]["hostname"] = str(hostname)

    hostname = cfg["auth"].get("hostname")

    if not hostname:
        dirty = True

        print("Robot setup (hit enter for default value):")
        while not hostname:
            hostname = input("Team number or robot hostname: ")

        cfg["auth"]["hostname"] = hostname

    if dirty:
        with open(cfg_filename, "w") as fp:
            cfg.write(fp)

    # see if an ssh alias exists
    try:
        with open(join(expanduser("~"), ".ssh", "config")) as fp:
            hn = hostname.lower()
            for line in fp:
                if re.match(r"\s*host\s+%s\s*" % hn, line.lower()):
                    no_resolve = True
                    break
    except Exception:
        pass

    # check to see if this is a team number
    team = None
    try:
        team = int(hostname.strip())
    except ValueError:
        # check to see if it matches a team hostname
        # -> allows legacy hostname configurations to benefit from
        #    the robot finder
        if not no_resolve:
            hostmod = hostname.lower().strip()
            m = re.search(r"10.(\d+).(\d+).2", hostmod)
            if m:
                team = int(m.group(1)) * 100 + int(m.group(2))
            else:
                m = re.match(r"roborio-(\d+)-frc(?:\.(?:local|lan))?$", hostmod)
                if m:
                    team = int(m.group(1))

    conn = None

    if team:
        logger.info("Finding robot for team %s", team)
        finder = RobotFinder(
            ("10.%d.%d.2" % (team // 100, team % 100), False),
            ("roboRIO-%d-FRC.local" % team, True),
            ("172.22.11.2", False),  # USB
            ("roboRIO-%d-FRC" % team, True),  # default DNS
            ("roboRIO-%d-FRC.lan" % team, True),
            ("roboRIO-%d-FRC.frc-field.local" % team, True),  # practice field mDNS
        )
        answer = finder.find()
        if not answer:
            raise Error("Could not find team %s robot" % team)

        no_resolve = True
        conn_hostname, conn = answer
    else:
        conn_hostname = hostname

    if not no_resolve:
        conn_hostname = _resolve_addr(hostname)

    logger.info("Connecting to robot via SSH at %s", conn_hostname)

    return SshController(conn_hostname, username, password, conn)



class Error(Exception):
    pass


class ArgError(Error):
    pass


class SshExecError(Error):
    def __init__(self, msg, retval):
        super().__init__(msg)
        self.retval = retval


class OpkgError(Exception):
    pass



import logging
import socket
import threading
import typing

from robotpy_installer.utils import _resolve_addr

logger = logging.getLogger("robotpy.installer")


class RobotFinder:
    def __init__(self, *addrs):
        self.tried = 0
        self.answer = None
        self.addrs = addrs
        self.cond = threading.Condition()

    def find(self) -> typing.Optional[typing.Tuple[str, socket.socket]]:
        with self.cond:
            self.tried = 0
            for addr, resolve in self.addrs:
                t = threading.Thread(target=self._try_server, args=(addr, resolve))
                t.setDaemon(True)
                t.start()

            while self.answer is None and self.tried != len(self.addrs):
                self.cond.wait()

            if self.answer:
                logger.info("-> Robot is at %s", self.answer[0])

            return self.answer

    def _try_server(self, addr: str, resolve: bool):
        success = False
        conn = None
        try:
            if resolve:
                addr = _resolve_addr(addr)

            conn = socket.create_connection((addr, 22), timeout=10)
            success = True
        except Exception:
            pass

        with self.cond:
            self.tried += 1
            if success and not self.answer:
                self.answer = (addr, conn)
            elif conn is not None:
                conn.close()

            self.cond.notify_all()



#
# This hacks around pip's lack of support for supporting environment markers
# when specifying --platform/--abi/--implementation etc
#

import platform
import runpy
import os
import sys


if __name__ == "__main__":
    # Setup environment for what the RoboRIO python would have
    # -> strictly speaking we only care about platform.machine as that's what
    #    we're using in robotpy-meta, but the rest for completeness
    platform.machine = lambda: "armv7l"
    platform.python_implementation = lambda: "CPython"
    platform.system = lambda: "Linux"
    platform.python_version = lambda: "3.11.0"

    runpy.run_module("pip", run_name="__main__")



import logging
import pathlib
import threading
from http.server import SimpleHTTPRequestHandler
from typing import Dict

from robotpy_installer.sshcontroller import SshController

logger = logging.getLogger("cacheserver")


class HTTPHandler(SimpleHTTPRequestHandler):
    def __init__(self, mapped_files, *args, **kwargs):
        self.mapped_files = mapped_files
        super().__init__(*args, **kwargs)

    def log_message(self, format: str, *args) -> None:
        logger.debug(f"%s {format}", self.address_string(), *args)

    def translate_path(self, path):
        xpath = path.split("?", 1)[0]
        xpath = xpath.split("#", 1)[0]
        redirect = self.mapped_files.get(xpath)
        if redirect:
            return redirect

        return super().translate_path(path)


class CacheServer:
    def __init__(self, ssh_controller: SshController, cache_root: pathlib.Path):
        self.controller = ssh_controller
        self.cache_root = cache_root

        self.transport = self.controller.client.get_transport()
        assert self.transport is not None
        self.port = self.transport.request_port_forward("", 0)

        self.mapped_files: Dict[str, str] = {}

    def add_mapping(self, fname: str, local_file: str):
        self.mapped_files[fname] = local_file

    def start(self):
        t = threading.Thread(target=self._handle_requests)
        t.setDaemon(True)
        t.start()

    def process_request(self, request):
        client_address = request.getpeername()
        try:
            HTTPHandler(
                self.mapped_files,
                request=request,
                client_address=client_address,
                server=None,
                directory=self.cache_root,
            ).handle()
        except OSError as e:
            if str(e) == "File is closed":
                return
        finally:
            request.close()

    def _handle_requests(self):
        request = self.transport.accept()

        while request is not None:
            t = threading.Thread(target=self.process_request, args=[request])
            t.setDaemon(True)
            t.start()

            request = self.transport.accept()



import pathlib
import sys
import subprocess


def test_download_basic():
    assert 0 == subprocess.check_call(
        [
            sys.executable,
            "-m",
            "robotpy_installer",
            "download",
            "-r",
            str(pathlib.Path(__file__).parent / "sample-requirements.txt"),
        ]
    )



import contextlib
import inspect
import logging
import pathlib
import re
import shutil
import subprocess
import sys
import typing

import click
from click import argument, option, group, pass_context, pass_obj, ClickException

from distutils.version import LooseVersion
from os.path import basename

from .version import version as __version__
from .cacheserver import CacheServer
from .errors import Error, SshExecError, OpkgError
from .opkgrepo import OpkgRepo
from .sshcontroller import SshController, ssh_from_cfg

_WPILIB_YEAR = "2023"
_IS_BETA = False

_OPKG_ARCH = "cortexa9-vfpv3"


_OPKG_FEEDS = [
    f"https://www.tortall.net/~robotpy/feeds/{_WPILIB_YEAR}",
    f"https://download.ni.com/ni-linux-rt/feeds/academic/2023/arm/main/{_OPKG_ARCH}",
    f"https://download.ni.com/ni-linux-rt/feeds/academic/2023/arm/extra/{_OPKG_ARCH}",
]

_ROBORIO_WHEELS = f"https://www.tortall.net/~robotpy/wheels/{_WPILIB_YEAR}/roborio"

_ROBORIO_IMAGES = [
    "2023_v3.1",
    "2023_v3.2",
]

_ROBORIO2_IMAGES = [
    "2023_v3.1",
    "2023_v3.2",
]

_ROBOTPY_PYTHON_PLATFORM = "linux_armv7l"
_ROBOTPY_PYTHON_VERSION_NUM = "311"
_ROBOTPY_PYTHON_VERSION = f"python{_ROBOTPY_PYTHON_VERSION_NUM}"


logger = logging.getLogger("robotpy.installer")


class RobotpyInstaller:
    def __init__(self, cache_root: pathlib.Path, cfgroot: pathlib.Path):
        self.cache_root = cache_root
        self.pip_cache = cache_root / "pip_cache"
        self.opkg_cache = cache_root / "opkg_cache"

        self.cfg_filename = cfgroot / ".installer_config"

    def log_startup(self) -> None:
        logger.info("RobotPy Installer %s", __version__)
        logger.info("-> caching files at %s", self.cache_root)

    def get_opkg(self, ssl_context) -> OpkgRepo:
        opkg = OpkgRepo(self.opkg_cache, _OPKG_ARCH, ssl_context)
        for feed in _OPKG_FEEDS:
            opkg.add_feed(feed)
        return opkg

    def get_opkg_packages(self, no_index: bool, ssl_context):
        opkg = self.get_opkg(ssl_context)
        if not no_index:
            opkg.update_packages()

        for feed in opkg.feeds:
            for _, pkgdata in feed.pkgs.items():
                for pkg in pkgdata:
                    yield pkg

    def get_ssh(self, robot: typing.Optional[str]) -> SshController:
        try:
            return ssh_from_cfg(
                self.cfg_filename, username="admin", password="", hostname=robot
            )
        except Error as e:
            raise ClickException(str(e)) from e

    def start_cache(self, ssh: SshController) -> CacheServer:
        cache = CacheServer(ssh, self.cache_root)
        cache.start()
        return cache


#
# Helpers
#


@contextlib.contextmanager
def catch_ssh_error(msg: str):
    try:
        yield
    except SshExecError as e:
        raise ClickException(f"{msg}: {e}")


def remove_legacy_components(ssh: SshController):
    # (remove in 2022) check for old robotpy components
    # -> only removes opkg components, pip will take care of the rest

    with catch_ssh_error("check for old RobotPy"):
        result = ssh.check_output("opkg list-installed python38*").strip()

    if result != "":
        packages = [line.split()[0] for line in result.splitlines()]

        print("RobotPy 2020 components detected!")
        for package in packages:
            print("-", package)

        if not click.confirm("Uninstall?"):
            raise ClickException("installer cannot continue")

        with catch_ssh_error("uninstall old RobotPy"):
            ssh.exec_cmd(f"opkg remove {' '.join(packages)}", print_output=True)


def show_disk_space(
    ssh: SshController,
) -> typing.Tuple[str, str, str]:
    with catch_ssh_error("checking free space"):
        result = ssh.check_output("df -h / | tail -n 1")

    _, size, used, _, pct, _ = result.strip().split()
    logger.info("-> RoboRIO disk usage %s/%s (%s full)", used, size, pct)

    return size, used, pct


def roborio_checks(
    ssh: SshController,
    ignore_image_version: bool,
    pip_check: bool = False,
):
    #
    # Image version check
    #

    with catch_ssh_error("retrieving image version"):
        result = ssh.check_output(
            "grep IMAGEVERSION /etc/natinst/share/scs_imagemetadata.ini",
        )

    roborio_match = re.match(r'IMAGEVERSION = "FRC_roboRIO_(.*)"', result.strip())
    roborio2_match = re.match(r'IMAGEVERSION = "FRC_roboRIO2_(.*)"', result.strip())

    if roborio_match:
        version = roborio_match.group(1)
        images = _ROBORIO_IMAGES
        name = "RoboRIO"
    elif roborio2_match:
        version = roborio2_match.group(1)
        images = _ROBORIO2_IMAGES
        name = "RoboRIO 2"
    else:
        version = "<unknown>"
        images = [
            f"({_ROBORIO_IMAGES[-1]} | {_ROBORIO2_IMAGES[-1]})",
        ]
        name = "RoboRIO (1 | 2)"

    logger.info(f"-> {name} image version: {version}")

    if not ignore_image_version and version not in images:
        raise ClickException(
            f"{name} image {images[-1]} is required! Use --ignore-image-version to install anyways"
        )

    #
    # Free space check.. maybe in the future we'll use this to not accidentally
    # fill the user's disk, but it'd be annoying to figure out
    #

    show_disk_space(ssh)

    #
    # Ensure that pip is installed
    #

    if pip_check:
        with catch_ssh_error("checking for pip3"):
            if ssh.exec_cmd("[ -x /usr/local/bin/pip3 ]").returncode != 0:
                raise ClickException(
                    inspect.cleandoc(
                        """
                        pip3 not found on RoboRIO, did you install python?

                        Use the 'download-python' and 'install-python' commands first!
                        """
                    )
                )


#
# Click-based CLI
#


def _make_ssl_context(use_certifi: bool):
    if not use_certifi:
        return None

    try:
        import certifi
    except ImportError:
        raise click.ClickException(
            "certifi is not installed, please install it via `pip install certifi`"
        )

    import ssl

    return ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=certifi.where())


@group()
@pass_context
def installer(ctx: click.Context):
    """
    RobotPy installer utility
    """

    log_datefmt = "%H:%M:%S"
    log_format = "%(asctime)s:%(msecs)03d %(levelname)-8s: %(name)-20s: %(message)s"

    logging.basicConfig(datefmt=log_datefmt, format=log_format, level=logging.INFO)

    cache_root = pathlib.Path.home() / "wpilib" / _WPILIB_YEAR / "robotpy"
    cfg_root = pathlib.Path.cwd()

    # This becomes the first argument to any cli command with the @pass_obj decorator
    ctx.obj = RobotpyInstaller(cache_root, cfg_root)


def _common_ssh_options(f):
    f = option("--robot", help="Specify the robot hostname or team number")(f)
    f = option(
        "--ignore-image-version", is_flag=True, help="Ignore RoboRIO image version"
    )(f)
    return f


#
# Cache management
#


@installer.group()
def cache():
    """Cache management"""


@cache.command()
@pass_obj
def location(installer: RobotpyInstaller):
    """Print cache location"""
    print(installer.cache_root)


@cache.command()
@option("-f", "--force", is_flag=True, help="Force removal without asking")
@pass_obj
def rm(installer: RobotpyInstaller, force: bool):
    """Delete all cached files"""
    if force or click.confirm(f"Really delete {installer.cache_root}?"):
        shutil.rmtree(installer.cache_root)


#
# OPkg related commands
#


@installer.group()
def opkg():
    """
    Advanced RoboRIO package management tools
    """


@opkg.command(name="download")
@option("--no-index", is_flag=True, help="Only examine local cache")
@option("--use-certifi", is_flag=True, help="Use SSL certificates from certifi")
@option(
    "-r",
    "--requirements",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help="Install from the given requirements file. This option can be used multiple times.",
)
@argument("packages", nargs=-1)
@pass_obj
def opkg_download(
    installer: RobotpyInstaller,
    no_index: bool,
    use_certifi: bool,
    requirements: typing.Tuple[str],
    packages: typing.Sequence[str],
):
    """
    Downloads opkg package to local cache
    """

    installer.log_startup()

    try:
        opkg = installer.get_opkg(_make_ssl_context(use_certifi))
        if not no_index:
            opkg.update_packages()

        if requirements:
            packages = list(packages) + opkg.load_opkg_from_req(*requirements)

        if not packages:
            raise ClickException("must specify packages to download")

        package_list = opkg.resolve_pkg_deps(packages)
        for package in package_list:
            opkg.download(package)
    except OpkgError as e:
        raise ClickException(str(e)) from e


@opkg.command(name="install")
@option(
    "--force-reinstall",
    is_flag=True,
    help="When upgrading, reinstall all packages even if they are already up-to-date.",
)
@option("--ignore-image-version", is_flag=True)
@option(
    "-r",
    "--requirements",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help="Install from the given requirements file. This option can be used multiple times.",
)
@argument("packages", nargs=-1, required=True)
@_common_ssh_options
@pass_obj
def opkg_install(
    installer: RobotpyInstaller,
    force_reinstall: bool,
    requirements: typing.Tuple[str],
    robot: str,
    ignore_image_version: bool,
    packages: typing.Sequence[str],
):
    """
    Installs opkg package on RoboRIO
    """

    installer.log_startup()

    opkg = installer.get_opkg(None)

    # Write out the install script
    # -> we use a script because opkg doesn't have a good mechanism
    #    to only install a package if it's not already installed
    opkg_files = []
    if requirements:
        packages = list(packages) + opkg.load_opkg_from_req(*requirements)

    try:
        packages = opkg.resolve_pkg_deps(packages)
    except OpkgError as e:
        raise ClickException(str(e))

    with installer.get_ssh(robot) as ssh:
        cache = installer.start_cache(ssh)

        roborio_checks(ssh, ignore_image_version)

        opkg_script = inspect.cleandoc(
            """
            set -e
            PACKAGES=()
            DO_INSTALL=0
            """
        )

        opkg_script_bit = inspect.cleandoc(
            f"""
            if ! opkg list-installed | grep -F "%(name)s - %(version)s"; then
                PACKAGES+=("http://localhost:{cache.port}/opkg_cache/%(fname)s")
                DO_INSTALL=1
            else
                echo "%(name)s already installed"
            fi
            """
        )

        for package in packages:
            try:
                pkg, fname = opkg.get_cached_pkg(package)
            except OpkgError as e:
                raise ClickException(str(e))

            opkg_script += "\n" + (
                opkg_script_bit
                % {
                    "fname": basename(fname),
                    "name": pkg["Package"],
                    "version": pkg["Version"],
                }
            )

            opkg_files.append(fname)

        # Finish it out
        opkg_script += "\n" + (
            inspect.cleandoc(
                """
                if [ "${DO_INSTALL}" == "0" ]; then
                    echo "No packages to install."
                else
                    echo + opkg install %(options)s ${PACKAGES[@]}
                    opkg install %(options)s ${PACKAGES[@]}
                fi

                sync
                ldconfig
                """
            )
            % {"options": "--force-reinstall" if force_reinstall else ""}
        )

        with catch_ssh_error("creating opkg install script"):
            # write to /tmp so that it doesn't persist
            ssh.exec_cmd(
                f"echo '{opkg_script}' > /tmp/install_opkg.sh",
                check=True,
            )

        with catch_ssh_error("installing selected packages"):
            ssh.exec_cmd("bash /tmp/install_opkg.sh", check=True, print_output=True)

        try:
            ssh.exec_cmd("rm /tmp/install_opkg.sh")
        except SshExecError:
            pass

        show_disk_space(ssh)


@opkg.command(name="list")
@option("--no-index", is_flag=True, help="Only examine local cache")
@option("--use-certifi", is_flag=True, help="Use SSL certificates from certifi")
@pass_obj
def opkg_list(installer: RobotpyInstaller, no_index: bool, use_certifi: bool):
    """
    List all packages in opkg database
    """

    data = set()
    for pkg in installer.get_opkg_packages(no_index, _make_ssl_context(use_certifi)):
        data.add("%(Package)s - %(Version)s" % pkg)

    for v in sorted(data):
        print(v)


@opkg.command(name="search")
@option("--no-index", is_flag=True, help="Only examine local cache")
@option("--use-certifi", is_flag=True, help="Use SSL certificates from certifi")
@argument("search")
@pass_obj
def opkg_search(
    installer: RobotpyInstaller, no_index: bool, use_certifi: bool, search: str
):
    """
    Search opkg database for packages
    """

    # TODO: make this more intelligent...
    data = set()
    for pkg in installer.get_opkg_packages(no_index, _make_ssl_context(use_certifi)):
        if search in pkg["Package"] or search in pkg.get("Description", ""):
            data.add("%(Package)s - %(Version)s" % pkg)
    for v in sorted(data):
        print(v)


@opkg.command(name="uninstall")
@argument("packages", nargs=-1, required=True)
@_common_ssh_options
@pass_obj
def opkg_uninstall(
    installer: RobotpyInstaller,
    robot: str,
    ignore_image_version: bool,
    packages: typing.Tuple[str],
):
    installer.log_startup()

    with installer.get_ssh(robot) as ssh:
        roborio_checks(ssh, ignore_image_version)

        package_list = " ".join(packages)

        with catch_ssh_error("removing packages"):
            ssh.exec_cmd(f"opkg remove {package_list}", check=True, print_output=True)

        show_disk_space(ssh)


#
# python installation
#


@installer.command()
@option("--use-certifi", is_flag=True, help="Use SSL certificates from certifi")
@pass_context
def download_python(ctx: click.Context, use_certifi: bool):
    """
    Downloads Python to a folder to be installed
    """
    ctx.forward(
        opkg_download, packages=[_ROBOTPY_PYTHON_VERSION], use_certifi=use_certifi
    )


@installer.command()
@_common_ssh_options
@pass_context
def install_python(
    ctx: click.Context,
    robot: str,
    ignore_image_version: bool,
):
    """
    Installs Python on a RoboRIO.

    Requires download-python to be executed first.
    """
    ctx.forward(opkg_install, packages=[_ROBOTPY_PYTHON_VERSION])


#
# Python package management
#


def _pip_options(f):
    f = option(
        "--force-reinstall",
        is_flag=True,
        help="When upgrading, reinstall all packages even if they are already up-to-date.",
    )(f)
    f = option(
        "--ignore-installed",
        "-I",
        is_flag=True,
        help="Ignore the installed packages (reinstalling instead)",
    )(f)
    f = option("--no-deps", is_flag=True, help="Don't install package dependencies")(f)
    f = option(
        "--pre",
        is_flag=True,
        default=_IS_BETA,
        help="Include pre-release and development versions",
    )(f)
    f = option(
        "--requirements",
        "-r",
        multiple=True,
        type=click.Path(exists=True),
        default=[],
        help="Install from the given requirements file. This option can be used multiple times.",
    )(f)

    f = argument("packages", nargs=-1)(f)
    return f


def _extend_pip_args(
    pip_args: typing.List[str],
    cache: typing.Optional[CacheServer],
    force_reinstall: bool,
    ignore_installed: bool,
    no_deps: bool,
    pre: bool,
    requirements: typing.Sequence[str],
):
    if pre:
        pip_args.append("--pre")
    if force_reinstall:
        pip_args.append("--force-reinstall")
    if ignore_installed:
        pip_args.append("--ignore-installed")
    if no_deps:
        pip_args.append("--no-deps")

    for req in requirements:
        if cache:
            fname = f"/requirements/{basename(req)}"
            cache.add_mapping(fname, req)
            pip_args.extend(["-r", f"http://localhost:{cache.port}{fname}"])
        else:
            pip_args.extend(["-r", req])


@installer.command()
@_pip_options
@pass_obj
def download(
    installer: RobotpyInstaller,
    force_reinstall: bool,
    ignore_installed: bool,
    no_deps: bool,
    pre: bool,
    requirements: typing.Tuple[str],
    packages: typing.Tuple[str],
):
    """
    Specify Python package(s) to download, and store them in the cache.

    You must be connected to the internet for this to work.
    """

    installer.log_startup()

    if not requirements and not packages:
        raise ClickException("You must give at least one requirement to download")

    try:
        import pip  # type: ignore
    except ImportError:
        raise ClickException("ERROR: pip must be installed to download python packages")

    try:
        pip_version = LooseVersion(pip.__version__)
    except:
        pass
    else:
        # TODO: what do we actually support? newer is better obviously...
        if pip_version < LooseVersion("18.0"):
            raise ClickException("robotpy-installer requires pip 18.0 or later")

    installer.pip_cache.mkdir(parents=True, exist_ok=True)

    pip_args = [
        "--no-cache-dir",
        "--disable-pip-version-check",
        "download",
        "--find-links",
        _ROBORIO_WHEELS,
        "--only-binary",
        ":all:",
        "--platform",
        _ROBOTPY_PYTHON_PLATFORM,
        "--python-version",
        _ROBOTPY_PYTHON_VERSION_NUM,
        "--implementation",
        "cp",
        "--abi",
        f"cp{_ROBOTPY_PYTHON_VERSION_NUM}",
        "-d",
        str(installer.pip_cache),
    ]

    _extend_pip_args(
        pip_args, None, force_reinstall, ignore_installed, no_deps, pre, requirements
    )

    pip_args.extend(packages)
    pip_args = [sys.executable, "-m", "robotpy_installer._pipstub"] + pip_args

    logger.debug("Using pip to download: %s", pip_args)

    retval = subprocess.call(pip_args)
    if retval != 0:
        raise ClickException("pip download failed")


@installer.command(name="install")
@_pip_options
@_common_ssh_options
@pass_obj
def pip_install(
    installer: RobotpyInstaller,
    force_reinstall: bool,
    ignore_installed: bool,
    no_deps: bool,
    pre: bool,
    requirements: typing.Tuple[str],
    packages: typing.Tuple[str],
    ignore_image_version: bool,
    robot: str,
):
    """
    Installs Python package(s) on a RoboRIO.

    The package must already been downloaded with the 'download' command first.
    """

    installer.log_startup()

    if len(requirements) == 0 and len(packages) == 0:
        raise ClickException("You must give at least one requirement to install")

    with installer.get_ssh(robot) as ssh:
        roborio_checks(ssh, ignore_image_version, pip_check=True)

        cachesvr = installer.start_cache(ssh)

        pip_args = [
            "/usr/local/bin/pip3",
            "--no-cache-dir",
            "--disable-pip-version-check",
            "install",
            "--no-index",
            "--find-links",
            f"http://localhost:{cachesvr.port}/pip_cache/",
            # always add --upgrade, anything in the cache should be installed
            "--upgrade",
            "--upgrade-strategy=eager",
        ]

        _extend_pip_args(
            pip_args,
            cachesvr,
            force_reinstall,
            ignore_installed,
            no_deps,
            pre,
            requirements,
        )

        pip_args.extend(packages)

        with catch_ssh_error("installing packages"):
            ssh.exec_cmd(" ".join(pip_args), check=True, print_output=True)

        # Some of our hacky wheels require this
        with catch_ssh_error("running ldconfig"):
            ssh.exec_cmd("ldconfig")

        show_disk_space(ssh)


@installer.command(name="list")
@_common_ssh_options
@pass_obj
def pip_list(
    installer: RobotpyInstaller,
    ignore_image_version: bool,
    robot: str,
):
    """
    Lists Python packages present on RoboRIO
    """
    installer.log_startup()

    with installer.get_ssh(robot) as ssh:
        roborio_checks(ssh, ignore_image_version, pip_check=True)

        with catch_ssh_error("pip3 list"):
            ssh.exec_cmd(
                "/usr/local/bin/pip3 --no-cache-dir --disable-pip-version-check list",
                check=True,
                print_output=True,
            )


@installer.command(name="uninstall")
@_common_ssh_options
@option(
    "--requirements",
    "-r",
    multiple=True,
    type=click.Path(exists=True),
    default=[],
    help="Install from the given requirements file. This option can be used multiple times.",
)
@argument("packages", nargs=-1)
@pass_obj
def pip_uninstall(
    installer: RobotpyInstaller,
    requirements: typing.Tuple[str],
    packages: typing.Tuple[str],
    ignore_image_version: bool,
    robot: str,
):
    """
    Uninstall Python packages from a RoboRIO
    """
    installer.log_startup()

    if len(requirements) == 0 and len(packages) == 0:
        raise ClickException("You must give at least one requirement to install")

    with installer.get_ssh(robot) as ssh:
        roborio_checks(ssh, ignore_image_version, pip_check=True)

        pip_args = [
            "/usr/local/bin/pip3",
            "--no-cache-dir",
            "--disable-pip-version-check",
            "uninstall",
            "--yes",
        ]
        pip_args.extend(packages)

        with catch_ssh_error("uninstalling packages"):
            ssh.exec_cmd(" ".join(pip_args), check=True, print_output=True)


#
# Removed commands
#


@installer.command(hidden=True)
def download_robotpy():
    raise ClickException(
        inspect.cleandoc(
            """

        The download-robotpy command has been removed! The equivalent commands are now:

            robotpy-installer download-python
            robotpy-installer download robotpy

        Run "robotpy-installer --help" for details.
            """
        )
    )


@installer.command(hidden=True)
def install_robotpy():
    raise ClickException(
        inspect.cleandoc(
            """

        The install-robotpy command has been removed! The equivalent commands are now:

            robotpy-installer install-python
            robotpy-installer install robotpy

        Run "robotpy-installer --help" for details.
            """
        )
    )


# alias for backwards compat
main = installer

if __name__ == "__main__":
    installer()
